"""
 Calculate the difference between two dictionaries as:
    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values

  Originally posted at http://stackoverflow.com/questions/1165352/fast-comparison-between-two-python-dictionary/1165552#1165552
  Available at repository: https://github.com/hughdbrown/dictdiffer

  Added the ability to recursively compare dictionaries
"""
import collections.abc as abc
import copy
from typing import Any
from typing import Dict
from typing import List
from typing import Set


def diff(current_dict, past_dict):
    return DictDiffer(current_dict, past_dict)


class DictDiffer:
    """
    Calculate the difference between two dictionaries as:
    (1) items added
    (2) items removed
    (3) keys same in both but changed values
    (4) keys same in both and unchanged values
    """

    def __init__(self, current_dict: Dict, past_dict: Dict):
        self.current_dict, self.past_dict = current_dict, past_dict
        self.set_current, self.set_past = set(list(current_dict)), set(list(past_dict))
        self.intersect = self.set_current.intersection(self.set_past)

    def added(self) -> Set:
        return self.set_current - self.intersect

    def removed(self) -> Set:
        return self.set_past - self.intersect

    def changed(self) -> Set:
        return {o for o in self.intersect if self.past_dict[o] != self.current_dict[o]}

    def unchanged(self) -> Set:
        return {o for o in self.intersect if self.past_dict[o] == self.current_dict[o]}


class RecursiveDictDiffer(DictDiffer):
    """
        Calculates a recursive diff between the current_dict and the past_dict
        creating a diff in the format

        {'new': new_value, 'old': old_value}

        It recursively searches differences in common keys whose values are
        dictionaries creating a diff dict in the format

        {'common_key' : {'new': new_value, 'old': old_value}

        The class overrides all DictDiffer methods, returning lists of keys and
        subkeys using the . notation (i.e 'common_key1.common_key2.changed_key')

        The class provides access to:
            (1) the added, removed, changes keys and subkeys (using the . notation)
                   ``added``, ``removed``, ``changed`` methods
            (2) the diffs in the format aboce (diff property)
                    ``diffs`` property
            (3) a dict with the new changed values only (new_values property)
                    ``new_values`` property
            (4) a dict with the old changed values only (old_values property)
                    ``old_values`` property
            (5) a string representation of the changes in the format:
                    ``changes_str`` property

        Note:
            The <_null_> value is a reserved value

    .. code-block:: text

                common_key1:
                  common_key2:
                    changed_key1 from '<old_str>' to '<new_str>'
                    changed_key2 from '[<old_elem1>, ..]' to '[<new_elem1>, ..]'
                common_key3:
                  changed_key3 from <old_int> to <new_int>

    """

    NONE_VALUE = "<_null_>"

    def __init__(self, past_dict: Dict, current_dict: Dict, ignore_missing_keys: bool):
        """
        past_dict
            Past dictionary.

        current_dict
            Current dictionary.

        ignore_missing_keys
            Flag specifying whether to ignore keys that no longer exist in the
            current_dict, but exist in the past_dict. If true, the diff will
            not contain the missing keys.
        """
        super().__init__(current_dict, past_dict)
        self._diffs = self._get_diffs(
            self.current_dict, self.past_dict, ignore_missing_keys
        )
        # Ignores unet values when assessing the changes
        self.ignore_unset_values = True

    @classmethod
    def _get_diffs(cls, dict1, dict2, ignore_missing_keys) -> Dict:
        """
        Returns a dict with the differences between dict1 and dict2

        Notes:
            Keys that only exist in dict2 are not included in the diff if
            ignore_missing_keys is True, otherwise they are
            Simple compares are done on lists
        """
        ret_dict = {}
        for p in dict1.keys():
            if p not in dict2:
                ret_dict.update({p: {"new": dict1[p], "old": cls.NONE_VALUE}})
            elif dict1[p] != dict2[p]:
                if isinstance(dict1[p], dict) and isinstance(dict2[p], dict):
                    sub_diff_dict = cls._get_diffs(
                        dict1[p], dict2[p], ignore_missing_keys
                    )
                    if sub_diff_dict:
                        ret_dict.update({p: sub_diff_dict})
                else:
                    ret_dict.update({p: {"new": dict1[p], "old": dict2[p]}})
        if not ignore_missing_keys:
            for p in dict2.keys():
                if p not in dict1.keys():
                    ret_dict.update({p: {"new": cls.NONE_VALUE, "old": dict2[p]}})
        return ret_dict

    @classmethod
    def _get_values(cls, diff_dict: Dict, type_: str = "new") -> Dict:
        """
        Returns a dictionaries with the 'new' values in a diff dict.

        type_
            Which values to return, 'new' or 'old'
        """
        ret_dict = {}
        for p in diff_dict.keys():
            if type_ in diff_dict[p].keys():
                ret_dict.update({p: diff_dict[p][type_]})
            else:
                ret_dict.update({p: cls._get_values(diff_dict[p], type_=type_)})
        return ret_dict

    @classmethod
    def _get_changes(cls, diff_dict: Dict) -> Dict:
        """
        Returns a list of string message with the differences in a diff dict.

        Each inner difference is tabulated two space deeper
        """
        changes_strings = []
        for p in sorted(diff_dict.keys()):
            if sorted(diff_dict[p].keys()) == ["new", "old"]:
                # Some string formatting
                old_value = diff_dict[p]["old"]
                if diff_dict[p]["old"] == cls.NONE_VALUE:
                    old_value = "nothing"
                elif isinstance(diff_dict[p]["old"], str):
                    old_value = "'{}'".format(diff_dict[p]["old"])
                elif isinstance(diff_dict[p]["old"], list):
                    old_value = "'{}'".format(", ".join(diff_dict[p]["old"]))
                new_value = diff_dict[p]["new"]
                if diff_dict[p]["new"] == cls.NONE_VALUE:
                    new_value = "nothing"
                elif isinstance(diff_dict[p]["new"], str):
                    new_value = "'{}'".format(diff_dict[p]["new"])
                elif isinstance(diff_dict[p]["new"], list):
                    new_value = "'{}'".format(", ".join(diff_dict[p]["new"]))
                changes_strings.append(f"{p} from {old_value} to {new_value}")
            else:
                sub_changes = cls._get_changes(diff_dict[p])
                if sub_changes:
                    changes_strings.append(f"{p}:")
                    changes_strings.extend([f"  {c}" for c in sub_changes])
        return changes_strings

    def added(self) -> Set[str]:
        """
        Returns all keys that have been added.

        If the keys are in child dictionaries they will be represented with
        . notation
        """

        def _added(diffs, prefix):
            keys = []
            for key in diffs.keys():
                if isinstance(diffs[key], dict) and "old" not in diffs[key]:
                    keys.extend(_added(diffs[key], prefix=f"{prefix}{key}."))
                elif diffs[key]["old"] == self.NONE_VALUE:
                    if isinstance(diffs[key]["new"], dict):
                        keys.extend(_added(diffs[key]["new"], prefix=f"{prefix}{key}."))
                    else:
                        keys.append(f"{prefix}{key}")
            return keys

        return sorted(_added(self._diffs, prefix=""))

    def removed(self) -> Set[str]:
        """
        Returns all keys that have been removed.

        If the keys are in child dictionaries they will be represented with
        . notation
        """

        def _removed(diffs, prefix):
            keys = []
            for key in diffs.keys():
                if isinstance(diffs[key], dict) and "old" not in diffs[key]:
                    keys.extend(_removed(diffs[key], prefix=f"{prefix}{key}."))
                elif diffs[key]["new"] == self.NONE_VALUE:
                    keys.append(f"{prefix}{key}")
                elif isinstance(diffs[key]["new"], dict):
                    keys.extend(_removed(diffs[key]["new"], prefix=f"{prefix}{key}."))
            return keys

        return sorted(_removed(self._diffs, prefix=""))

    def changed(self) -> Set[str]:
        """
        Returns all keys that have been changed.

        If the keys are in child dictionaries they will be represented with
        . notation
        """

        def _changed(diffs, prefix):
            keys = []
            for key in diffs.keys():
                if not isinstance(diffs[key], dict):
                    continue

                if isinstance(diffs[key], dict) and "old" not in diffs[key]:
                    keys.extend(_changed(diffs[key], prefix=f"{prefix}{key}."))
                    continue
                if self.ignore_unset_values:
                    if (
                        "old" in diffs[key]
                        and "new" in diffs[key]
                        and diffs[key]["old"] != self.NONE_VALUE
                        and diffs[key]["new"] != self.NONE_VALUE
                    ):
                        if isinstance(diffs[key]["new"], dict):
                            keys.extend(
                                _changed(
                                    diffs[key]["new"],
                                    prefix=f"{prefix}{key}.",
                                )
                            )
                        else:
                            keys.append(f"{prefix}{key}")
                    elif isinstance(diffs[key], dict):
                        keys.extend(_changed(diffs[key], prefix=f"{prefix}{key}."))
                else:
                    if "old" in diffs[key] and "new" in diffs[key]:
                        if isinstance(diffs[key]["new"], dict):
                            keys.extend(
                                _changed(
                                    diffs[key]["new"],
                                    prefix=f"{prefix}{key}.",
                                )
                            )
                        else:
                            keys.append(f"{prefix}{key}")
                    elif isinstance(diffs[key], dict):
                        keys.extend(_changed(diffs[key], prefix=f"{prefix}{key}."))

            return keys

        return sorted(_changed(self._diffs, prefix=""))

    def unchanged(self) -> Set[str]:
        """
        Returns all keys that have been unchanged.

        If the keys are in child dictionaries they will be represented with
        . notation
        """

        def _unchanged(current_dict, diffs, prefix):
            keys = []
            for key in current_dict.keys():
                if key not in diffs:
                    keys.append(f"{prefix}{key}")
                elif isinstance(current_dict[key], dict):
                    if "new" in diffs[key]:
                        # There is a diff
                        continue
                    else:
                        keys.extend(
                            _unchanged(
                                current_dict[key],
                                diffs[key],
                                prefix=f"{prefix}{key}.",
                            )
                        )

            return keys

        return sorted(_unchanged(self.current_dict, self._diffs, prefix=""))

    @property
    def diffs(self) -> Dict:
        """Returns a dict with the recursive diffs current_dict - past_dict"""
        return self._diffs

    @property
    def new_values(self) -> Dict:
        """Returns a dictionary with the new values"""
        return self._get_values(self._diffs, type_="new")

    @property
    def old_values(self) -> Dict:
        """Returns a dictionary with the old values"""
        return self._get_values(self._diffs, type_="old")

    @property
    def changes_str(self) -> str:
        """Returns a string describing the changes"""
        return "\n".join(self._get_changes(self._diffs))


def deep_diff(old: Dict, new: Dict, ignore: List = None) -> Dict[str, Any]:
    ignore = ignore or []
    res = {}
    old = copy.deepcopy(old) or {}
    new = copy.deepcopy(new) or {}

    stack = [(old, new, False)]

    while len(stack) > 0:
        tmps = []
        tmp_old, tmp_new, reentrant = stack.pop()
        for key in set(list(tmp_old) + list(tmp_new)):
            if key in tmp_old and key in tmp_new and tmp_old[key] == tmp_new[key]:
                del tmp_old[key]
                del tmp_new[key]
                continue
            if not reentrant:
                if key in tmp_old and key in ignore:
                    del tmp_old[key]
                if key in tmp_new and key in ignore:
                    del tmp_new[key]
                if isinstance(tmp_old.get(key), abc.Mapping) and isinstance(
                    tmp_new.get(key), abc.Mapping
                ):
                    tmps.append((tmp_old[key], tmp_new[key], False))
        if tmps:
            stack.extend([(tmp_old, tmp_new, True)] + tmps)
    if old:
        res["old"] = old
    if new:
        res["new"] = new
    return res


def recursive_diff(
    past_dict: Dict, current_dict: Dict, ignore_missing_keys: bool = True
) -> RecursiveDictDiffer:
    """
    Returns a RecursiveDictDiffer object that computes the recursive diffs
    between two dictionaries

    past_dict
            Past dictionary

    current_dict
        Current dictionary

    ignore_missing_keys
        Flag specifying whether to ignore keys that no longer exist in the
        current_dict, but exist in the past_dict. If true, the diff will
        not contain the missing keys.
        Default is True.
    """
    return RecursiveDictDiffer(past_dict, current_dict, ignore_missing_keys)



import copy
import fnmatch
import logging
import pathlib
import re
from collections.abc import Mapping
from collections.abc import MutableMapping
from collections.abc import Sequence
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Set

import aiofiles
import msgpack
import yaml

from . import args
from . import DEFAULT_TARGET_DELIM

log = logging.getLogger(__name__)


class CaseInsensitiveDict(MutableMapping):
    """
    Inspired by requests' case-insensitive dict implementation, but works with
    non-string keys as well.
    """

    def __init__(self, init=None, **kwargs):
        """
        Force internal dict to be ordered to ensure a consistent iteration
        order, irrespective of case.
        """
        self._data = {}
        self.update(init or {}, **kwargs)

    def __len__(self):
        return len(self._data)

    def __setitem__(self, key, value):
        # Store the case-sensitive key so it is available for dict iteration
        self._data[to_lowercase(key)] = (key, value)

    def __delitem__(self, key):
        del self._data[to_lowercase(key)]

    def __getitem__(self, key):
        return self._data[to_lowercase(key)][1]

    def __iter__(self):
        return (item[0] for item in self._data.values())

    def __eq__(self, rval):
        if not isinstance(rval, Mapping):
            # Comparing to non-mapping type (e.g. int) is always False
            return False
        return dict(self.items_lower()) == dict(CaseInsensitiveDict(rval).items_lower())

    def __repr__(self):
        return repr(dict(self.items()))

    def items_lower(self):
        """
        Returns a generator iterating over keys and values, with the keys all
        being lowercase.
        """
        return ((key, val[1]) for key, val in self._data.items())

    def copy(self):
        """
        Returns a copy of the object
        """
        return CaseInsensitiveDict(self._data.items())


class ImmutableDict(Mapping):
    """
    An abstract base class that implements the interface of a `dict` but is immutable.
    Items can be retrieved via namespacing.
    No values can be changed after initialization
    """

    def __init__(self, init_: Dict[str, Any], **c_kwargs):
        """
        :param init_: A dictionary from which to inherit data
        """
        init_.update(**c_kwargs)
        values = {}
        for k, v in init_.items():
            if isinstance(v, Dict):
                values[k] = ImmutableDict(init_=v)
            elif isinstance(v, (tuple, int, str, bytes)):
                values[k] = v
            elif isinstance(v, Iterable):
                values[k] = tuple(v)
            else:
                values[k] = v
        # __setattr__ is borked (on purpose) so we have to call it from super() right here
        super().__setattr__("_ImmutableDict__store", values)

    def __setattr__(self, k: str, v: Any):
        raise TypeError(
            f"{self.__class__.__name__} does not support attribute assignment"
        )

    def __getattr__(self, k: str):
        return self.__store[k]

    def __getitem__(self, k: str) -> Any:
        return self.__store[k]

    def __contains__(self, k: str) -> bool:
        return k in self.__store

    def __iter__(self):
        return iter(self.__store)

    def __len__(self) -> int:
        return len(self.__store.keys())

    def __copy__(self) -> Dict[str, Any]:
        ret = {}
        # Unpack IMAP items so that it's turtles all the way down
        for k, v in self.__store.items():
            if isinstance(v, ImmutableDict):
                ret[k] = v.__copy__()
            else:
                ret[k] = v
        return ret

    def __repr__(self):
        return repr(copy.copy(self))


class NamespaceDict(dict):
    """
    A dictionary that can access it's string keys through the namespace
    """

    def __init__(self, seq: Mapping = None, **kwargs):
        """
        NamespaceDict() -> new empty namespaced dictionary
        NamespaceDict(mapping) -> new namespaced dictionary initialized from a mapping object's
            (key, value) pairs
        NamespaceDict(iterable) -> new namespaced dictionary initialized as if via:
            d = {}
            for k, v in iterable:
                d[k] = v
        NamespaceDict(**kwargs) -> new namespaced dictionary initialized with the name=value pairs
            in the keyword argument list.  For example:  NamespaceDict(one=1, two=2)
        """
        if seq:
            if isinstance(seq, Mapping):
                super().__init__()
                for k, v in seq.items():
                    self[k] = v
            else:
                super().__init__(seq)
        else:
            super().__init__()

        for k, v in kwargs.items():
            self[k] = v

    def __setitem__(self, k, v):
        if isinstance(v, Mapping) and not isinstance(v, NamespaceDict):
            v = NamespaceDict(v)
            self[k] = v
        else:
            super().__setitem__(k, v)

    def __setattr__(self, k: str, v: Any):
        if isinstance(v, Mapping) and not isinstance(v, NamespaceDict):
            v = NamespaceDict(v)
        self[k] = v

    def __getattr__(self, k: str):
        if k.startswith("_"):
            return super().__getattribute__(k)
        return self[k]

    def __copy__(self):
        return {k: copy.copy(v) for k, v in self.items()}

    def __deepcopy__(self, memodict=None):
        if memodict is None:
            memodict = {}
        return NamespaceDict(copy.deepcopy(self.copy(), memodict))


class SafeNamespaceDict(NamespaceDict):
    """
    A Dict that does not expose values when represented as a string
    """

    def __repr__(self):
        return str({k: "*" for k in self})

    def __str__(self):
        return repr(self)


class FileDict(Mapping):
    """
    A dictionary type that stores it's information in a file
    """

    def __init__(self, cache: str, seq: Iterable = None, **kwargs):
        self.cache = pathlib.Path(cache)
        if seq:
            for k in seq:
                self[k] = seq[k]
        for k, v in kwargs.items():
            self[k] = v

    def get(self, k, default=None):
        if self.cache.exists():
            with open(self.cache, "rb+") as fh_:
                return msgpack.loads(fh_.read()).get(k, default)
        else:
            return default

    async def aget(self, k, default=None):
        if self.cache.exists():
            async with aiofiles.open(self.cache, "rb+") as fh_:
                return msgpack.loads(await fh_.read()).get(k, default)
        else:
            return default

    async def aset(self, k, v):
        if self.cache.exists():
            async with aiofiles.open(self.cache, "rb+") as fh_:
                data = msgpack.loads(await fh_.read())
        else:
            data = {}
        data[k] = v
        async with aiofiles.open(self.cache, "wb+") as fh_:
            await fh_.write(msgpack.dumps(data))

    def __setitem__(self, k, v):
        if self.cache.exists():
            with open(self.cache, "rb+") as fh_:
                data = msgpack.loads(fh_.read())[k]
        else:
            data = {}
        data[k] = v
        with open(self.cache, "wb+") as fh_:
            fh_.write(msgpack.dumps(data))

    def __getitem__(self, k):
        with open(self.cache, "rb+") as fh_:
            return msgpack.loads(fh_.read())[k]

    async def __aiter__(self):
        async with aiofiles.open(self.cache, "rb+") as fh_:
            for k in msgpack.loads(await fh_.read()):
                yield k

    def __iter__(self) -> Iterator:
        with open(self.cache, "rb+") as fh_:
            return iter(msgpack.loads(fh_))

    def __len__(self) -> int:
        with open(self.cache, "rb+") as fh_:
            return len(msgpack.loads(fh_.read()))


def __change_case(data, attr, preserve_dict_class=False):
    """
    Calls data.attr() if data has an attribute/method called attr.
    Processes data recursively if data is a Mapping or Sequence.
    For Mapping, processes both keys and values.
    """
    try:
        return getattr(data, attr)()
    except AttributeError:
        pass

    data_type = data.__class__

    if isinstance(data, Mapping):
        return (data_type if preserve_dict_class else dict)(
            (
                __change_case(key, attr, preserve_dict_class),
                __change_case(val, attr, preserve_dict_class),
            )
            for key, val in data.items()
        )
    if isinstance(data, Sequence):
        return data_type(
            __change_case(item, attr, preserve_dict_class) for item in data
        )
    return data


def _remove_circular_refs(ob, _seen: Set = None):
    """
    Generic method to remove circular references from objects.
    This has been taken from author Martijn Pieters
    https://stackoverflow.com/questions/44777369/
    remove-circular-references-in-dicts-lists-tuples/44777477#44777477
    :param ob: dict, list, typle, set, and frozenset
        Standard python object
    :param object _seen:
        Object that has circular reference
    :returns:
        Cleaned Python object
    """
    if _seen is None:
        _seen = set()
    if id(ob) in _seen:
        # Here we caught a circular reference.
        # Alert user and cleanup to continue.
        log.exception(
            "Caught a circular reference in data structure below."
            "Cleaning and continuing execution.\n%r\n",
            ob,
        )
        return None
    _seen.add(id(ob))
    res = ob
    if isinstance(ob, dict):
        res = {
            _remove_circular_refs(k, _seen): _remove_circular_refs(v, _seen)
            for k, v in ob.items()
        }
    elif isinstance(ob, (list, tuple, set, frozenset)):
        res = type(ob)(_remove_circular_refs(v, _seen) for v in ob)
    # remove id again; only *nested* references count
    _seen.remove(id(ob))
    return res


def compare_dicts(old: Dict = None, new: Dict = None) -> Dict[str, Dict]:
    """
    Compare before and after results from various salt functions, returning a
    dict describing the changes that were made.
    """
    ret = {}
    for key in set(new or {}).union(old or {}):
        if key not in old:
            # New key
            ret[key] = {"old": "", "new": new[key]}
        elif key not in new:
            # Key removed
            ret[key] = {"new": "", "old": old[key]}
        elif new[key] != old[key]:
            # Key modified
            ret[key] = {"old": old[key], "new": new[key]}
    return ret


def object_to_dict(obj) -> Dict:
    """
    Convert an object to a dictionary
    """
    if isinstance(obj, list) or isinstance(obj, tuple):
        ret = []
        for item in obj:
            ret.append(object_to_dict(item))
    elif hasattr(obj, "__dict__"):
        ret = {}
        for item in obj.__dict__:
            if item.startswith("_"):
                continue
            ret[item] = object_to_dict(obj.__dict__[item])
    else:
        ret = obj
    return ret


def is_dictlist(data: List) -> bool:
    """
    Returns True if data is a list of one-element dicts (as found in many SLS
    schemas), otherwise returns False
    """
    if isinstance(data, list):
        for element in data:
            if isinstance(element, dict):
                if len(element) != 1:
                    return False
            else:
                return False
        return True
    return False


def recursive_diff(
    old: Iterable,
    new: Iterable,
    ignore_keys: List = None,
    ignore_order: bool = False,
    ignore_missing_keys: bool = False,
) -> Dict[str, Iterable]:
    """
    Performs a recursive diff on mappings and/or iterables and returns the result
    in a {'old': values, 'new': values}-style.
    Compares dicts and sets unordered (obviously), OrderedDicts and Lists ordered
    (but only if both ``old`` and ``new`` are of the same type),
    all other Mapping types unordered, and all other iterables ordered.

    :param old: Mapping or Iterable to compare from.
    :param new: Mapping or Iterable to compare to.
    :param ignore_keys: List of keys to ignore when comparing Mappings.
    :param ignore_order: Compare ordered mapping/iterables as if they were unordered.
    :param ignore_missing_keys: Do not return keys only present in ``old``
        but missing in ``new``. Only works for regular dicts.
    :return dict: Returns dict with keys 'old' and 'new' containing the differences.
    """
    ignore_keys = ignore_keys or []
    ret_old = copy.deepcopy(old)
    ret_new = copy.deepcopy(new)
    if isinstance(old, Mapping) and isinstance(new, Mapping) and not ignore_order:
        append_old, append_new = [], []
        if len(old) != len(new):
            min_length = min(len(old), len(new))
            # The list coercion is required for Py3
            append_old = list(old.keys())[min_length:]
            append_new = list(new.keys())[min_length:]
        # Compare ordered
        for (key_old, key_new) in zip(old, new):
            if key_old == key_new:
                if key_old in ignore_keys:
                    del ret_old[key_old]
                    del ret_new[key_new]
                else:
                    res = recursive_diff(
                        old[key_old],
                        new[key_new],
                        ignore_keys=ignore_keys,
                        ignore_order=ignore_order,
                        ignore_missing_keys=ignore_missing_keys,
                    )
                    if not res:  # Equal
                        del ret_old[key_old]
                        del ret_new[key_new]
                    else:
                        ret_old[key_old] = res["old"]
                        ret_new[key_new] = res["new"]
            else:
                if key_old in ignore_keys:
                    del ret_old[key_old]
                if key_new in ignore_keys:
                    del ret_new[key_new]
        # If the OrderedDicts were of inequal length, add the remaining key/values.
        for item in append_old:
            ret_old[item] = old[item]
        for item in append_new:
            ret_new[item] = new[item]
        ret = {"old": ret_old, "new": ret_new} if ret_old or ret_new else {}
    elif isinstance(old, Mapping) and isinstance(new, Mapping):
        # Compare unordered
        for key in set(list(old) + list(new)):
            if key in ignore_keys:
                del ret_old[key]
                del ret_new[key]
            elif key in old and key in new:
                res = recursive_diff(
                    old[key],
                    new[key],
                    ignore_keys=ignore_keys,
                    ignore_order=ignore_order,
                    ignore_missing_keys=ignore_missing_keys,
                )
                if not res:  # Equal
                    del ret_old[key]
                    del ret_new[key]
                else:
                    ret_old[key] = res["old"]
                    ret_new[key] = res["new"]
            elif ignore_missing_keys and key in old:
                del ret_old[key]
        ret = {"old": ret_old, "new": ret_new} if ret_old or ret_new else {}
    elif isinstance(old, set) and isinstance(new, set):
        ret = {"old": old - new, "new": new - old} if old - new or new - old else {}
    elif (
        isinstance(old, Iterable)
        and not isinstance(old, str)
        and isinstance(new, Iterable)
        and not isinstance(new, str)
    ):
        # Create a list so we can edit on an index-basis.
        list_old = list(ret_old)
        list_new = list(ret_new)
        if ignore_order:
            for item_old in old:
                for item_new in new:
                    res = recursive_diff(
                        item_old,
                        item_new,
                        ignore_keys=ignore_keys,
                        ignore_order=ignore_order,
                        ignore_missing_keys=ignore_missing_keys,
                    )
                    if not res:
                        list_old.remove(item_old)
                        list_new.remove(item_new)
                        continue
        else:
            remove_indices = []
            for index, (iter_old, iter_new) in enumerate(zip(old, new)):
                res = recursive_diff(
                    iter_old,
                    iter_new,
                    ignore_keys=ignore_keys,
                    ignore_order=ignore_order,
                    ignore_missing_keys=ignore_missing_keys,
                )
                if not res:  # Equal
                    remove_indices.append(index)
                else:
                    list_old[index] = res["old"]
                    list_new[index] = res["new"]
            for index in reversed(remove_indices):
                list_old.pop(index)
                list_new.pop(index)
        # Instantiate a new whatever-it-was using the list as iterable source.
        # This may not be the most optimized in way of speed and memory usage,
        # but it will work for all iterable types.
        ret = (
            {"old": type(old)(list_old), "new": type(new)(list_new)}
            if list_old or list_new
            else {}
        )
    else:
        ret = {} if old == new else {"old": ret_old, "new": ret_new}
    return ret


def repack_dictlist(data, strict=False, recurse=False, key_cb=None, val_cb=None):
    """
    Takes a list of one-element dicts (as found in many SLS schemas) and
    repacks into a single dictionary.
    """
    if isinstance(data, str):
        try:
            data = yaml.safe_load(data)
        except yaml.parser.ParserError as err:
            log.error(err)
            return {}

    if key_cb is None:
        key_cb = lambda x: x
    if val_cb is None:
        val_cb = lambda x, y: y

    valid_non_dict = (str, int, float)
    if isinstance(data, list):
        for element in data:
            if isinstance(element, valid_non_dict):
                continue
            if isinstance(element, dict):
                if len(element) != 1:
                    log.error(
                        "Invalid input for repack_dictlist: key/value pairs "
                        "must contain only one element (data passed: %s).",
                        element,
                    )
                    return {}
            else:
                log.error(
                    "Invalid input for repack_dictlist: element %s is "
                    "not a string/dict/numeric value",
                    element,
                )
                return {}
    else:
        log.error(
            "Invalid input for repack_dictlist, data passed is not a list " "(%s)", data
        )
        return {}

    ret = {}
    for element in data:
        if isinstance(element, valid_non_dict):
            ret[key_cb(element)] = None
        else:
            key = next(iter(element))
            val = element[key]
            if is_dictlist(val):
                if recurse:
                    ret[key_cb(key)] = repack_dictlist(val, recurse=recurse)
                elif strict:
                    log.error(
                        "Invalid input for repack_dictlist: nested dictlist "
                        "found, but recurse is set to False"
                    )
                    return {}
                else:
                    ret[key_cb(key)] = val_cb(key, val)
            else:
                ret[key_cb(key)] = val_cb(key, val)
    return ret


def subdict_match(
    data: Dict,
    expr: str,
    delimiter: str = DEFAULT_TARGET_DELIM,
    regex_match: bool = False,
    exact_match: bool = False,
) -> bool:
    """
    Check for a match in a dictionary using a delimiter character to denote
    levels of subdicts, and also allowing the delimiter character to be
    matched. Thus, 'foo:bar:baz' will match data['foo'] == 'bar:baz' and
    data['foo']['bar'] == 'baz'. The latter would take priority over the
    former, as more deeply-nested matches are tried first.
    """

    def _match(target, pattern):
        target = str(target).lower()
        pattern = str(pattern).lower()

        if regex_match:
            try:
                return re.match(pattern, target)
            except Exception:  # pylint: disable=broad-except
                log.error("Invalid regex '%s' in match", pattern)
                return False
        else:
            return (
                target == pattern if exact_match else fnmatch.fnmatch(target, pattern)
            )

    def _dict_match(target, pattern):
        ret = False
        wildcard = pattern.startswith("*:")
        if wildcard:
            pattern = pattern[2:]

        if pattern == "*":
            # We are just checking that the key exists
            ret = True
        if not ret and pattern in target:
            # We might want to search for a key
            ret = True
        if not ret and subdict_match(
            target, pattern, regex_match=regex_match, exact_match=exact_match
        ):
            ret = True
        if not ret and wildcard:
            for key in target:
                if isinstance(target[key], dict):
                    if _dict_match(
                        target[key],
                        pattern,
                    ):
                        return True
                elif isinstance(target[key], list):
                    for item in target[key]:
                        if _match(
                            item,
                            pattern,
                        ):
                            return True
                elif _match(
                    target[key],
                    pattern,
                ):
                    return True
        return ret

    splits = expr.split(delimiter)
    num_splits = len(splits)
    if num_splits == 1:
        # Delimiter not present, this can't possibly be a match
        return False

    # If we have 4 splits, then we have three delimiters. Thus, the indexes we
    # want to use are 3, 2, and 1, in that order.
    for idx in range(num_splits - 1, 0, -1):
        key = delimiter.join(splits[:idx])
        if key == "*":
            # We are matching on everything under the top level, so we need to
            # treat the match as the entire data being passed in
            matchstr = expr
            match = data
        else:
            matchstr = delimiter.join(splits[idx:])
            match = traverse_dict_and_list(data, key, {}, delimiter=delimiter)
        log.debug(
            "Attempting to match '%s' in '%s' using delimiter '%s'",
            matchstr,
            key,
            delimiter,
        )
        if match == {}:
            continue
        if isinstance(match, dict):
            if _dict_match(match, matchstr):
                return True
            continue
        if isinstance(match, (list, tuple)):
            # We are matching a single component to a single list member
            for member in match:
                if isinstance(member, dict):
                    if _dict_match(
                        member,
                        matchstr,
                    ):
                        return True
                if _match(member, matchstr):
                    return True
            continue
        if _match(match, matchstr):
            return True
    return False


def to_lowercase(data, preserve_dict_class=False):
    """
    Recursively changes everything in data to lowercase.
    """
    return __change_case(data, "lower", preserve_dict_class)


def to_uppercase(data, preserve_dict_class=False):
    """
    Recursively changes everything in data to uppercase.
    """
    return __change_case(data, "upper", preserve_dict_class)


def traverse_dict(
    data: Dict, key: str, default: Any = None, delimiter: str = DEFAULT_TARGET_DELIM
):
    """
    Traverse a dict using a colon-delimited (or otherwise delimited, using the
    'delimiter' param) target string. The target 'foo:bar:baz' will return
    data['foo']['bar']['baz'] if this value exists, and will otherwise return
    the dict in the default argument.
    """
    ptr = data
    try:
        for each in key.split(delimiter):
            ptr = ptr[each]
    except (KeyError, IndexError, TypeError):
        # Encountered a non-indexable value in the middle of traversing
        return default
    return ptr


def traverse_dict_and_list(
    data: Dict or List,
    key: Any,
    default: Any = None,
    delimiter: str = DEFAULT_TARGET_DELIM,
):
    """
    Traverse a dict or list using a colon-delimited (or otherwise delimited,
    using the 'delimiter' param) target string. The target 'foo:bar:0' will
    return data['foo']['bar'][0] if this value exists, and will otherwise
    return the dict in the default argument.
    Function will automatically determine the target type.
    The target 'foo:bar:0' will return data['foo']['bar'][0] if data like
    {'foo':{'bar':['baz']}} , if data like {'foo':{'bar':{'0':'baz'}}}
    then return data['foo']['bar']['0']
    """
    ptr = data
    for each in key.split(delimiter):
        if isinstance(ptr, list):
            try:
                idx = int(each)
            except ValueError:
                embed_match = False
                # Index was not numeric, lets look at any embedded dicts
                for embedded in (x for x in ptr if isinstance(x, dict)):
                    try:
                        ptr = embedded[each]
                        embed_match = True
                        break
                    except KeyError:
                        pass
                if not embed_match:
                    # No embedded dicts matched, return the default
                    return default
            else:
                try:
                    ptr = ptr[idx]
                except IndexError:
                    return default
        else:
            try:
                ptr = ptr[each]
            except KeyError:
                # YAML-load the current key (catches integer/float dict keys)
                try:
                    loaded_key = args.yamlify_arg(each)
                except Exception:  # pylint: disable=broad-except
                    return default
                if loaded_key == each:
                    # After YAML-loading, the desired key is unchanged. This
                    # means that the KeyError caught above is a legitimate
                    # failure to match the desired key. Therefore, return the
                    # default.
                    return default
                else:
                    # YAML-loading the key changed its value, so re-check with
                    # the loaded key. This is how we can match a numeric key
                    # with a string-based expression.
                    try:
                        ptr = ptr[loaded_key]
                    except (KeyError, TypeError):
                        return default
            except TypeError:
                return default
    return ptr



"""
    This library makes it possible to introspect dataset and aggregate nodes
    when it is instructed.

    .. note::

        The following examples with be expressed in YAML for convenience's sake:

        - !aggr-scalar will refer to Scalar python function
        - !aggr-map will refer to Map python object
        - !aggr-seq will refer for Sequence python object


    How to instructs merging
    ------------------------

    This yaml document has duplicate keys:

    .. code-block:: yaml

        foo: !aggr-scalar first
        foo: !aggr-scalar second
        bar: !aggr-map {first: foo}
        bar: !aggr-map {second: bar}
        baz: !aggr-scalar 42

    but tagged values instruct Salt that overlapping values they can be merged
    together:

    .. code-block:: yaml

        foo: !aggr-seq [first, second]
        bar: !aggr-map {first: foo, second: bar}
        baz: !aggr-seq [42]


    Default merge strategy is keep untouched
    ----------------------------------------

    For example, this yaml document still has duplicate keys, but does not
    instruct aggregation:

    .. code-block:: yaml

        foo: first
        foo: second
        bar: {first: foo}
        bar: {second: bar}
        baz: 42

    So the late found values prevail:

    .. code-block:: yaml

        foo: second
        bar: {second: bar}
        baz: 42


    Limitations
    -----------

    Aggregation is permitted between tagged objects that share the same type.
    If not, the default merge strategy prevails.

    For example, these examples:

    .. code-block:: yaml

        foo: {first: value}
        foo: !aggr-map {second: value}

        bar: !aggr-map {first: value}
        bar: 42

        baz: !aggr-seq [42]
        baz: [fail]

        qux: 42
        qux: !aggr-scalar fail

    are interpreted like this:

    .. code-block:: yaml

        foo: !aggr-map{second: value}

        bar: 42

        baz: [fail]

        qux: !aggr-seq [fail]


    Introspection
    -------------

    TODO: write this part

"""
import collections
import copy
import logging
from typing import Iterable
from typing import Tuple

__all__ = ["aggregate", "Aggregate", "Map", "Scalar", "Sequence"]

log = logging.getLogger(__name__)


class Aggregate:
    """
    Aggregation base.
    """


class Map(collections.OrderedDict, Aggregate):
    """
    Map aggregation.
    """


class Sequence(list, Aggregate):
    """
    Sequence aggregation.
    """


def Scalar(obj):
    """
    Shortcut for Sequence creation

    >>> Scalar('foo') == Sequence(['foo'])
    True
    """
    return Sequence([obj])


def levelise(level: bool or int or Iterable) -> Tuple[bool, bool or int]:
    """
    Describe which levels are allowed to do deep merging.

    level can be:

    True
        all levels are True

    False
        all levels are False

    an int
        only the first levels are True, the others are False

    a sequence
        it describes which levels are True, it can be:

        * a list of bool and int values
        * a string of 0 and 1 characters

    """

    if not level:  # False, 0, [] ...
        return False, False
    if level is True:
        return True, True
    if isinstance(level, int):
        return True, level - 1
    try:  # a sequence
        deep, subs = int(level[0]), level[1:]
        return bool(deep), subs
    except Exception as error:  # pylint: disable=broad-except
        log.warning(error)
        raise


def mark(
    obj: object, map_class: object = Map, sequence_class: object = Sequence
) -> object:
    """
    Convert obj into an Aggregate instance
    """
    if isinstance(obj, Aggregate):
        return obj
    if isinstance(obj, dict):
        return map_class(obj)
    if isinstance(obj, (list, tuple, set)):
        return sequence_class(obj)
    else:
        return sequence_class([obj])


def aggregate(
    obj_a,
    obj_b,
    level: bool or int = False,
    map_class: object = Map,
    sequence_class: object = Sequence,
):
    """
    Merge obj_b into obj_a.

    >>> aggregate('first', 'second', True) == ['first', 'second']
    True
    """
    deep, subdeep = levelise(level)

    if deep:
        obj_a = mark(obj_a, map_class=map_class, sequence_class=sequence_class)
        obj_b = mark(obj_b, map_class=map_class, sequence_class=sequence_class)

    if isinstance(obj_a, dict) and isinstance(obj_b, dict):
        if isinstance(obj_a, Aggregate) and isinstance(obj_b, Aggregate):
            # deep merging is more or less a.update(obj_b)
            response = copy.copy(obj_a)
        else:
            # introspection on obj_b keys only
            response = copy.copy(obj_b)

        for key, value in obj_b.items():
            if key in obj_a:
                value = aggregate(obj_a[key], value, subdeep, map_class, sequence_class)
            response[key] = value
        return response

    if isinstance(obj_a, Sequence) and isinstance(obj_b, Sequence):
        response = obj_a.__class__(obj_a[:])
        for value in obj_b:
            if value not in obj_a:
                response.append(value)
        return response

    response = copy.copy(obj_b)

    if isinstance(obj_a, Aggregate) or isinstance(obj_b, Aggregate):
        log.info("only one value marked as aggregate. keep `obj_b` value")
        return response

    log.debug("no value marked as aggregate. keep `obj_b` value")
    return response



from typing import Type
from typing import Union


class _SpecialForm:
    __slots__ = ("_name", "__doc__", "_getitem")

    def __init__(self, getitem):
        self._getitem = getitem
        self._name = getitem.__name__
        self.__doc__ = getitem.__doc__

    def __getitem__(self, parameters):
        return self._getitem(self, parameters)


# Types for type-hinting, they cannot be used in casting
@_SpecialForm
def Computed(self, type_: Type):
    """
    The "Computed" type indicates that a variable is the result of internal computation.
    It is not used by any function and effects no change.
    The variable's value is associated with the function soley for reference.

    Example:

    .. code-block:: python

      from dict_tools import typing


      def function(arg1: typing.Computed[str]):
          ...
    """
    return Union[type_, type(None)]


@_SpecialForm
def Sensitive(self, type_: Type):
    """
    The "Sensitive" type indicates that a variable contains data that should be protected from being displayed to users.

    Example:

    .. code-block:: python

      from dict_tools import typing


      def function(arg1: typing.Sensitive[str]):
          ...
    """
    return Union[type_, type(None)]



from typing import ByteString
from typing import Dict
from typing import Iterable
from typing import Mapping


def decode(value):
    if isinstance(value, ByteString):
        return value.decode()
    elif isinstance(value, str):
        return value
    elif isinstance(value, (Mapping, Dict)):
        return decode_dict(value)
    elif isinstance(value, Iterable):
        return value.__class__(decode(x) for x in value)
    else:
        return value


def decode_dict(data: Dict[bytes, bytes]) -> Dict[str, str]:
    """
    Recursively decode all byte-strings found in a dictionary
    """
    return {decode(key): decode(value) for key, value in data.items()}
